"""
If snapshots need to be (re)generated, set GENERATE_SNAPSHOTS=1
and run the tests as usual.
"""

import asyncio
import os
import platform
import subprocess
from typing import Callable
from typing import Dict
from typing import Generator

import aiohttp
from ddtrace._trace.sampler import DatadogSampler
from ddtrace._trace.sampling_rule import SamplingRule
from ddtrace.profiling import Profiler
from ddtrace.propagation.http import HTTPPropagator
from ddtrace.trace import Tracer
import pytest


@pytest.fixture
def tracer(testagent_port: str, testagent: aiohttp.ClientSession) -> Tracer:
    tracer = Tracer()
    tracer._agent_url = f"http://localhost:{testagent_port}"
    return tracer


@pytest.fixture
def trace_sample_rate() -> float:
    return 1.0


@pytest.fixture
def stats_tracer(tracer: Tracer, trace_sample_rate: float) -> Generator[Tracer, None, None]:
    tracer.configure(compute_stats_enabled=True)
    tracer._sampler = DatadogSampler(rules=[SamplingRule(sample_rate=trace_sample_rate)])
    for processor in tracer._span_processors:
        if processor.__class__.__name__ == "SpanStatsProcessorV06":
            processor._agent_url = tracer._agent_url
    yield tracer


@pytest.mark.parametrize(
    "operation_name,service,resource,error,span_type,meta,metrics,response_code",
    [
        # First value is the reference data (also stored in the snapshot)
        ("root", "custom_service", "/url/endpoint", 0, "web", {}, {}, 200),
        ("root2", "custom_service", "/url/endpoint", 0, "web", {}, {}, 400),
        ("root", "custom_service2", "/url/endpoint", 0, "web", {}, {}, 400),
        ("root", "custom_service", "/url/endpoint/2", 0, "web", {}, {}, 400),
        ("root", "custom_service", "/url/endpoint", 1, "web", {}, {}, 400),
        ("root", "custom_service", "/url/endpoint", 0, "http", {}, {}, 400),
        (
            "root",
            "custom_service",
            "/url/endpoint",
            0,
            "http",
            {"meta": "value"},
            {},
            400,
        ),
        (
            "root",
            "custom_service",
            "/url/endpoint",
            0,
            "http",
            {},
            {"metrics": 2.3},
            400,
        ),
        ("root", "custom_service", "/url/endpoint", 0, "web", {}, {}, 200),
    ],
)
async def test_single_trace(
    testagent,
    testagent_url,
    tracer,
    operation_name,
    service,
    resource,
    error,
    span_type,
    meta,
    metrics,
    response_code,
):
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_single_trace")
    tracer = Tracer()
    tracer._agent_url = testagent_url
    with tracer.trace(operation_name, service=service, resource=resource, span_type=span_type) as span:
        if error is not None:
            span.error = error
        for k, v in meta.items():
            span.set_tag(k, v)
        for k, v in metrics.items():
            span.set_metric(k, v)
    tracer.shutdown()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_single_trace")

    assert resp.status == response_code


async def test_multi_trace(testagent_url, testagent, tracer):
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_multi_trace")
    with tracer.trace("root0"):
        with tracer.trace("child0"):
            pass
    with tracer.trace("root1"):
        with tracer.trace("child1"):
            pass
    tracer.flush()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_multi_trace")
    assert resp.status == 200

    # Run the snapshot test again.
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_multi_trace")
    with tracer.trace("root0"):
        with tracer.trace("child0"):
            pass
    with tracer.trace("root1"):
        with tracer.trace("child1"):
            pass
    tracer.flush()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_multi_trace")
    assert resp.status == 200

    # Simulate a failed snapshot with a missing trace.
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_multi_trace")
    with tracer.trace("root0"):
        with tracer.trace("child0"):
            pass
    tracer.flush()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_multi_trace")
    assert resp.status == 400
    tracer.shutdown()


async def test_trace_distributed_same_payload(testagent_url, testagent, tracer):
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_trace_distributed_same_payload")
    with tracer.trace("root0"):
        with tracer.trace("child0") as span:
            ctx = span.context

    tracer.context_provider.activate(ctx)
    with tracer.trace("root1"):
        with tracer.trace("child1"):
            pass
    tracer.flush()
    resp = await testagent.get(
        f"{testagent_url}/test/session/snapshot?test_session_token=test_trace_distributed_same_payload"
    )
    assert resp.status == 200


async def test_trace_distributed_propagated(testagent_url, testagent, tracer):
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_trace_distributed_propagated")
    headers = {
        "x-datadog-trace-id": "1234",
        "x-datadog-parent-id": "5678",
    }
    context = HTTPPropagator.extract(headers)
    tracer.context_provider.activate(context)

    with tracer.trace("root"):
        with tracer.trace("child"):
            pass
    tracer.flush()
    resp = await testagent.get(
        f"{testagent_url}/test/session/snapshot?test_session_token=test_trace_distributed_propagated"
    )

    assert resp.status == 200, await resp.text()


async def test_trace_missing_received(testagent_url, testagent, tracer):
    resp = await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_trace_missing_received")
    assert resp.status == 200, await resp.text()

    with tracer.trace("root0"):
        with tracer.trace("child0"):
            pass
    tracer.flush()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_trace_missing_received")
    assert resp.status == 200

    # Do another snapshot without sending any traces.
    resp = await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_trace_missing_received")
    assert resp.status == 200, await resp.text()
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_trace_missing_received")
    assert resp.status == 400


def _tracestats_traces(tracer: Tracer) -> None:
    for i in range(5):
        with tracer.trace("http.request", resource="/users/view") as span:
            if i == 4:
                span.error = 1


def _tracestats_traces_no_error(tracer: Tracer) -> None:
    for i in range(5):
        with tracer.trace("http.request", resource="/users/view"):
            pass


def _tracestats_traces_missing_trace(tracer: Tracer) -> None:
    for i in range(4):
        with tracer.trace("http.request", resource="/users/view") as span:
            if i == 3:
                span.error = 1


def _tracestats_traces_extra_trace(tracer: Tracer) -> None:
    _tracestats_traces(tracer)
    with tracer.trace("http.request", resource="/users/list"):
        pass


@pytest.mark.parametrize("trace_sample_rate", [0.0])  # Don't send any traces
@pytest.mark.parametrize(
    "do_traces,fail",
    [
        (
            _tracestats_traces,
            False,
        ),  # Keep this parametrization first as the next ones depend on it.
        (_tracestats_traces_no_error, True),
        (_tracestats_traces_missing_trace, True),
        (_tracestats_traces_extra_trace, True),
    ],
)
async def test_tracestats(
    testagent_url: str,
    testagent: aiohttp.ClientSession,
    stats_tracer: Tracer,
    testagent_snapshot_ci_mode: bool,
    trace_sample_rate: float,
    do_traces: Callable[[Tracer], None],
    fail: bool,
) -> None:
    await testagent.get(f"{testagent_url}/test/session/start?test_session_token=test_trace_stats")
    do_traces(stats_tracer)
    stats_tracer.shutdown()  # force out the stats
    resp = await testagent.get(f"{testagent_url}/test/session/snapshot?test_session_token=test_trace_stats")
    if fail:
        assert resp.status == 400
    else:
        assert resp.status == 200, await resp.text()


async def test_cmd(testagent_url: str, testagent: aiohttp.ClientSession, tracer: Tracer) -> None:
    """Test the commands provided with the library.

    Note that this test reuses the trace/snapshot from test_single_trace above.
    """
    env = os.environ.copy()
    env["DD_TRACE_AGENT_URL"] = testagent_url
    p = subprocess.run(
        ["ddapm-test-agent-session-start", "--test-session-token=test_single_trace"],
        env=env,
    )
    assert p.returncode == 0
    with tracer.trace("root", service="custom_service", resource="/url/endpoint", span_type="web"):
        pass
    tracer.flush()
    p = subprocess.run(["ddapm-test-agent-snapshot", "--test-session-token=test_single_trace"], env=env)
    assert p.returncode == 0

    # Ensure failing snapshots work.
    p = subprocess.run(
        ["ddapm-test-agent-session-start", "--test-session-token=test_single_trace"],
        env=env,
    )
    assert p.returncode == 0
    with tracer.trace("root1234", service="custom_service", resource="/url/endpoint", span_type="web"):
        pass
    tracer.shutdown()
    p = subprocess.run(["ddapm-test-agent-snapshot", "--test-session-token=test_single_trace"], env=env)
    assert p.returncode == 1


@pytest.mark.skipif(platform.system() == "Windows", reason="Profiling broken on windows")
async def test_profiling_endpoint(testagent_url: str, testagent: aiohttp.ClientSession, tracer: Tracer) -> None:
    p = Profiler(tracer=tracer)
    p.start()
    p.stop(flush=True)
    resp = await testagent.get(f"{testagent_url}/test/session/requests")
    assert resp.status == 200
    data = await resp.json()
    assert len(data) >= 1
    assert data[-1]["url"].endswith("/profiling/v1/input")


async def test_race_condition(
    testagent: aiohttp.ClientSession,
    testagent_port: int,
    v04_reference_http_trace_payload_headers: Dict[str, str],
    v04_reference_http_trace_payload_data: bytes,
    testagent_url: str,
) -> None:
    """
    Reproduction of: "RuntimeError: readany() called while another coroutine is already waiting for incoming data"
    when trace requests are made and being read simultaneously
    """
    reqs = []
    for i in range(50):
        reqs.append(
            testagent.put(
                testagent_url + "/v0.4/traces",
                headers=v04_reference_http_trace_payload_headers,
                data=v04_reference_http_trace_payload_data,
            )
        )
        reqs.append(testagent.get(testagent_url + "/test/session/traces"))

    resps = await asyncio.gather(*reqs)
    for r in resps:
        assert r.status == 200



import asyncio
import json
import logging
from typing import Dict
from typing import List

from aiohttp.web import Request
from multidict import CIMultiDictProxy

from .checks import Check
from .trace import Span


log = logging.getLogger(__name__)


class CheckTraceCountHeader(Check):
    name = "trace_count_header"
    description = """
The number of traces included in a payload must be included as the
X-Datadog-Trace-Count http header with each payload. The value of the
header must match the number of traces included in the payload.
""".strip()

    def check(self, headers: CIMultiDictProxy[str], num_traces: int) -> None:
        if "X-Datadog-Trace-Count" not in headers:
            self.fail("X-Datadog-Trace-Count header not found in headers")
            return
        try:
            count = int(headers["X-Datadog-Trace-Count"])
        except ValueError:
            self.fail("X-Datadog-Trace-Count header is not a valid integer")
            return
        else:
            if num_traces != count:
                self.fail(
                    f"X-Datadog-Trace-Count value ({count}) does not match actual number of traces ({num_traces})"
                )


class CheckMetaTracerVersionHeader(Check):
    name = "meta_tracer_version_header"
    description = """v0.4 payloads must include the Datadog-Meta-Tracer-Version header."""

    def check(self, headers: CIMultiDictProxy[str]) -> None:
        if "Datadog-Meta-Tracer-Version" not in headers:
            self.fail("Datadog-Meta-Tracer-Version not found in headers")


class CheckTraceContentLength(Check):
    name = "trace_content_length"
    description = """
The max content size of a trace payload is 50MB.
""".strip()

    def check(self, headers: CIMultiDictProxy[str]) -> None:
        if "Content-Length" not in headers:
            self.fail(f"content length header 'Content-Length' not in http headers {headers}")
            return
        content_length = int(headers["Content-Length"])
        if content_length > 5e7:
            self.fail(f"content length {content_length} too large.")


class CheckTraceStallAsync(Check):
    name = "trace_stall"
    description = """
Stall the trace (mimicking an overwhelmed or throttled agent) for the given duration in seconds.

Enable the check by submitting the X-Datadog-Test-Stall-Seconds http header (unit is seconds)
with the request.

Note that only the request for this trace is stalled, subsequent requests will not be
affected.
""".strip()

    async def check(self, headers: CIMultiDictProxy[str], request: Request) -> None:
        if "X-Datadog-Test-Stall-Seconds" in headers:
            duration = float(headers["X-Datadog-Test-Stall-Seconds"])
        else:
            duration = request.app["trace_request_delay"]
        if duration > 0:
            log.info("Stalling for %r seconds.", duration)
            await asyncio.sleep(duration)


class CheckTracePeerService(Check):
    name = "trace_peer_service"
    description = """
The ``peer.service`` tag is correctly set for Client / Producer spans.
""".strip()
    default_enabled = True

    def check(self, span: Span, dd_config_env: Dict[str, str]) -> None:
        log.info("Performing ``peer.service`` Span Check")
        meta = span.get("meta", {})

        whitelisted_components = ["couchbase"]

        if dd_config_env.get("DD_TRACE_SPAN_ATTRIBUTE_SCHEMA", "v0") != "v0":
            if "peer.service" in meta.keys():
                for component in whitelisted_components:
                    if component in meta.get("component", ""):
                        skipped_component = meta.get("component", "")
                        log.debug(
                            f"Skipped ``peer.service`` Span Check for Span: {span['name']} with component {skipped_component}."
                        )
                        self.skip(
                            f"Skipped ``peer.service`` Span Check for Span: {span['name']} with component {skipped_component}."
                        )
                        return

                # if meta.get("peer.service", None) is None:
                # self.fail(json.dumps(span, indent=4) + f"\nSpan: {span['name']} of kind: {meta['span.kind']} should have tag ``peer.service`` set.")

                peer_service = meta.get("peer.service")
                peer_service_source_key = meta.get("_dd.peer.service.source", "")
                peer_service_source_val = meta.get(peer_service_source_key, "")
                if peer_service != peer_service_source_val:
                    self.fail(
                        json.dumps(span, indent=4)
                        + f"\nSpan: {span['name']} expected to have ``peer.service`` tag equal to ``{peer_service_source_key}`` of: {peer_service_source_val}, actual: {peer_service}."
                    )
                log.debug(f"Successfully completed `peer.`service`` tag Span Check for Span: {span['name']}")
                return
            else:
                log.debug(f"Skipped ``peer.service`` Span Check for Span: {span['name']} with no `peer.service` tag")
                self.skip(f"Skipped ``peer.service`` Span Check for Span: {span['name']} with no `peer.service` tag")
        else:
            log.debug(
                f"Skipped ``peer.service`` Span Check for Span: {span['name']} with DD_TRACE_SPAN_ATTRIBUTE_SCHEMA `v0`."
            )
            self.skip(
                f"Skipped ``peer.service`` Span Check for Span: {span['name']} with DD_TRACE_SPAN_ATTRIBUTE_SCHEMA `v0`."
            )


class CheckTraceDDService(Check):
    name = "trace_dd_service"
    description = """
The ``service`` name is correctly set to ``DD_SERVICE`` for V1 auto-instrumented spans.
""".strip()
    default_enabled = True

    def check(self, trace: List[Span], dd_config_env: Dict[str, str]) -> None:
        log.info("Performing ``DD_SERVICE`` Trace Check")

        # trace context can be set to service for each span
        trace_context = None

        whitelisted_components = ["rabbitmq", "jms", "kafka", "java-web-servlet-response"]

        if dd_config_env.get("DD_TRACE_SPAN_ATTRIBUTE_SCHEMA", "v0") != "v1":
            log.debug("Skipping Span Check `trace_dd_service` for Span Attribute Schema v0")
            self.skip("Skipping Span Check `trace_dd_service` for Span Attribute Schema v0")
            return

        if dd_config_env.get("DD_TRACE_HTTP_CLIENT_SPLIT_BY_DOMAIN", False) or dd_config_env.get(
            "DD_TRACE_DB_CLIENT_SPLIT_BY_INSTANCE", False
        ):
            log.debug(
                f"Skipped ``DD_SERVICE`` Span Check for trace with SPLIT_BY config variable set within config: \n   {json.dumps(dd_config_env, indent=4)}."
            )
            self.skip(
                f"Skipped ``DD_SERVICE`` Span Check for trace with SPLIT_BY config variable set within config: \n   {json.dumps(dd_config_env, indent=4)}."
            )
            return

        for span in trace:
            meta = span.get("meta", {})
            for component in whitelisted_components:
                if component in meta.get("component", ""):
                    skipped_component = meta.get("component", "")
                    log.debug(
                        f"Skipped ``DD_SERVICE`` Span Check for Span: {span['name']} with component {skipped_component}."
                    )
                    self.skip(
                        f"Skipped ``DD_SERVICE`` Span Check for Span: {span['name']} with component {skipped_component}."
                    )
                    return

            component = meta.get("component", "")
            if component != "":
                dd_service = dd_config_env.get(f"DD_{component.upper()}_SERVICE", None)
                if not dd_service:
                    log.debug("DD_SERVICE not set for component: %s. Args %s", component, dd_config_env)
                dd_service = dd_service or dd_config_env.get("DD_SERVICE", None)
                if dd_service is None:
                    self.fail(
                        json.dumps(dd_config_env, indent=4)
                        + f"\n``DD_SERVICE`` env not set for Span: {span['name']} with service: {span['service']}."
                    )

                service = span.get("service")
                if service != dd_service:
                    # check for special case where span is of type web and has context as service
                    if "servlet.context" in meta.keys() or trace_context:
                        trace_context = (
                            trace_context
                            if trace_context is not None
                            else meta.get("servlet.context", "").replace("/", "")
                        )
                        if service != trace_context:
                            self.fail(
                                json.dumps(span, indent=4)
                                + f"\nSpan: {span['name']} expected to have ``service`` name equal to context of ``{trace_context}``. Actual: {service}."
                            )
                        elif span.get("type") in ["web", "http"]:
                            log.debug(
                                f"Skipped ``DD_SERVICE`` Span Check for Span: {span['name']} of type: [`web`, `http`]"
                            )
                            self.skip(
                                f"Skipped ``DD_SERVICE`` Span Check for Span: {span['name']} of type: [`web`, `http`]"
                            )
                            pass
                    else:
                        self.fail(
                            json.dumps(span, indent=4)
                            + f"\nSpan: {span['name']} expected to have ``service`` name equal to DD_SERVICE of ``{dd_service}``. Actual: {service}."
                        )
                else:
                    log.debug(f"Successfully completed ``service`` name Span Check for Span: {span['name']}")
        return



import asyncio
import json
import os
from pathlib import Path
import random
import socket
import subprocess
from typing import Awaitable
from typing import Dict
from typing import Generator
from typing import List
from typing import Literal
from typing import Optional
from typing import Set
from typing import cast
from urllib.parse import urlparse

import aiohttp
from aiohttp import web
from aiohttp.client_exceptions import ClientConnectorError
from aiohttp.client_exceptions import ClientOSError
from aiohttp.web import Response
from ddsketch import LogCollapsingLowestDenseDDSketch
from ddsketch.pb.proto import DDSketchProto
from ddtrace import config
import grpc.aio as grpc_aio
import msgpack
from opentelemetry.proto.collector.logs.v1.logs_service_pb2_grpc import LogsServiceStub
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2_grpc import MetricsServiceStub
import pytest

from ddapm_test_agent.agent import DEFAULT_OTLP_GRPC_PORT
from ddapm_test_agent.agent import DEFAULT_OTLP_HTTP_PORT
from ddapm_test_agent.agent import _parse_csv
from ddapm_test_agent.agent import make_app
from ddapm_test_agent.agent import make_otlp_grpc_server_async
from ddapm_test_agent.apmtelemetry import TelemetryEvent
from ddapm_test_agent.client import TestOTLPClient
from ddapm_test_agent.logs import LOGS_ENDPOINT
from ddapm_test_agent.metrics import METRICS_ENDPOINT
from ddapm_test_agent.trace import Span
from ddapm_test_agent.trace import Trace
from ddapm_test_agent.trace_snapshot import DEFAULT_SNAPSHOT_IGNORES


# Fix the service name to make tests consistently pass local and in CI.
config.service = ""

pytest_plugins = "aiohttp.pytest_plugin"


@pytest.fixture
def agent_enabled_checks() -> Generator[List[str], None, None]:
    yield ["trace_content_length", "meta_tracer_version_header", "trace_count_header", "trace_stall"]


@pytest.fixture
def log_span_fmt() -> Generator[str, None, None]:
    yield "[{name}]"


@pytest.fixture
def snapshot_dir(tmp_path: Path) -> Generator[Path, None, None]:
    yield tmp_path


@pytest.fixture
def snapshot_ci_mode() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def snapshot_ignored_attrs() -> Generator[Set[str], None, None]:
    yield set(_parse_csv(DEFAULT_SNAPSHOT_IGNORES))


@pytest.fixture
def agent_url() -> Generator[str, None, None]:
    yield ""


@pytest.fixture
def trace_request_delay() -> Generator[float, None, None]:
    yield 0.0


@pytest.fixture
def suppress_trace_parse_errors() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def pool_trace_check_failures() -> Generator[bool, None, None]:
    yield True


@pytest.fixture
def disable_error_responses() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def snapshot_removed_attrs() -> Generator[Set[str], None, None]:
    yield set()


@pytest.fixture
def snapshot_regex_placeholders() -> Generator[Dict[str, str], None, None]:
    yield dict()


@pytest.fixture
def vcr_cassettes_directory() -> Generator[str, None, None]:
    import shutil
    import tempfile

    vcr_dir = tempfile.mkdtemp(prefix="vcr-cassettes-")
    try:
        yield vcr_dir
    finally:
        shutil.rmtree(vcr_dir, ignore_errors=True)


@pytest.fixture
def vcr_ci_mode() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def vcr_provider_map() -> Generator[str, None, None]:
    yield ""


@pytest.fixture
def vcr_ignore_headers() -> Generator[str, None, None]:
    yield ""


@pytest.fixture
def dd_site() -> Generator[str, None, None]:
    yield "datadoghq.com"


@pytest.fixture
def dd_api_key() -> Generator[str, None, None]:
    yield "1234567890"


@pytest.fixture
def disable_llmobs_data_forwarding() -> Generator[bool, None, None]:
    yield True


@pytest.fixture
async def agent_app(
    aiohttp_server,
    agent_enabled_checks,
    log_span_fmt,
    snapshot_dir,
    snapshot_ci_mode,
    snapshot_ignored_attrs,
    agent_url,
    trace_request_delay,
    suppress_trace_parse_errors,
    pool_trace_check_failures,
    disable_error_responses,
    snapshot_removed_attrs,
    snapshot_regex_placeholders,
    vcr_cassettes_directory,
    vcr_ci_mode,
    vcr_provider_map,
    vcr_ignore_headers,
    dd_site,
    dd_api_key,
    disable_llmobs_data_forwarding,
):
    app = await aiohttp_server(
        make_app(
            enabled_checks=agent_enabled_checks,
            log_span_fmt=log_span_fmt,
            snapshot_dir=str(snapshot_dir),
            snapshot_ci_mode=snapshot_ci_mode,
            snapshot_ignored_attrs=snapshot_ignored_attrs,
            agent_url=agent_url,
            trace_request_delay=trace_request_delay,
            suppress_trace_parse_errors=suppress_trace_parse_errors,
            pool_trace_check_failures=pool_trace_check_failures,
            disable_error_responses=disable_error_responses,
            snapshot_removed_attrs=snapshot_removed_attrs,
            snapshot_regex_placeholders=snapshot_regex_placeholders,
            vcr_cassettes_directory=vcr_cassettes_directory,
            vcr_ci_mode=vcr_ci_mode,
            vcr_provider_map=vcr_provider_map,
            vcr_ignore_headers=vcr_ignore_headers,
            dd_site=dd_site,
            dd_api_key=dd_api_key,
            disable_llmobs_data_forwarding=disable_llmobs_data_forwarding,
        )
    )
    yield app


@pytest.fixture
async def agent(agent_app, aiohttp_client, loop):
    client = await aiohttp_client(agent_app)
    yield client


@pytest.fixture
def v1_reference_http_trace_payload_data() -> bytes:
    data = msgpack.packb(
        {
            11: [
                {  # chunk
                    1: 0,  # Priority
                    6: bytes(
                        [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0xE3]
                    ),  # Trace ID 128 bits (85, 8675)
                    4: [
                        {  # Span
                            1: "my-service",
                            2: "span-name",
                            3: "GET /res",
                            4: 12345678,
                            9: [
                                "foo",
                                1,
                                "bar",
                                "foo2",
                                1,
                                5,
                                "some-num",
                                4,
                                42,
                            ],
                        }
                    ],
                }
            ]
        }
    )
    return cast(bytes, data)


@pytest.fixture
def v04_reference_http_trace_payload_data_raw() -> List[Trace]:
    data = [
        [
            Span(
                {
                    "name": "http.request",
                    "service": "my-http-server",
                    "trace_id": random.randint(0, 2**64),
                    "span_id": random.randint(0, 2**64),
                    "parent_id": None,
                    "resource": "/users/",
                    "type": "http",
                    "start": 1342343123,
                    "duration": 123214,
                    "error": 0,
                    "meta": {
                        "component": "",
                        "http.url": "http://localhost:8080/users",
                        "http.method": "GET",
                        "http.status_code": "200",
                        "http.status_msg": "OK",
                        "language": "python",
                        "runtime-id": "2d377516ca12429aaf072f037ed2e4cc",
                    },
                    "metrics": {
                        "process_id": 111,
                        "sampling_priority_v1": 1.0,
                    },
                    "span_events": [
                        {
                            "name": "event1",
                            "time_unix_nano": 123,
                            "attributes": {
                                "string": {"type": 0, "string_value": "foo"},
                                "array": {
                                    "type": 4,
                                    "array_value": {
                                        "values": [{"type": 2, "int_value": 1}, {"type": 2, "int_value": 2}],
                                    },
                                },
                            },
                        },
                    ],
                }
            ),
        ]
    ]
    return data


@pytest.fixture
def v04_reference_http_trace_payload_data(
    v04_reference_http_trace_payload_data_raw: List[Trace],
) -> bytes:
    return cast(bytes, msgpack.packb(v04_reference_http_trace_payload_data_raw))


@pytest.fixture
def v04_reference_http_trace_payload_headers() -> Dict[str, str]:
    headers = {
        "Content-Type": "application/msgpack",
        "X-Datadog-Trace-Count": "1",
        "Datadog-Meta-Tracer-Version": "v0.1",
        "datadog-meta-lang": "python",
        "X-Datadog-Trace-Env-Variables": "DD_INTEGRATION=express,DD_INTEGRATION_VERSION=1.2.3",
    }
    return headers


def v04_trace(  # type: ignore
    agent,
    traces: List[Trace],
    encoding: Literal["msgpack", "json"] = "msgpack",
    token: Optional[str] = None,
):
    params = {"test_session_token": token} if token is not None else {}
    if encoding == "msgpack":
        content_type = "application/msgpack"
        encode = msgpack.packb
    else:
        content_type = "application/json"
        encode = json.dumps

    headers = {
        "Content-Type": content_type,
        "X-Datadog-Trace-Count": str(len(traces)),
        "Datadog-Meta-Tracer-Version": "v0.1",
    }

    return agent.put(
        "/v0.4/traces",
        params=params,
        headers=headers,
        data=encode(traces),
    )


@pytest.fixture
def do_reference_v04_http_trace(
    agent,
    v04_reference_http_trace_payload_headers,
    v04_reference_http_trace_payload_data,
):
    def fn(token: Optional[str] = None, payload_override: Optional[bytes] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.put(  # type: ignore
            "/v0.4/traces",
            params=params,
            headers=v04_reference_http_trace_payload_headers,
            data=payload_override or v04_reference_http_trace_payload_data,
        )

    yield fn


@pytest.fixture
def v06_reference_http_stats_payload_headers():
    headers = {
        "Content-Type": "application/msgpack",
        "Datadog-Meta-Lang": "python",
        "Datadog-Meta-Tracer-Version": "v0.1",
    }
    yield headers


@pytest.fixture
def v06_reference_http_stats_payload_data_raw():
    ok_dist = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)
    err_dist = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)

    rng = random.Random(0)
    total = 0
    ok_n = 97
    err_n = 3
    for _ in range(ok_n):
        n = rng.randint(int(1e9), int(2e9))
        total += n
        ok_dist.add(n)
    for _ in range(err_n):
        n = rng.randint(int(2e9), int(3e9))
        total += n
        err_dist.add(n)

    data = {
        "Env": "dev",
        "Version": "v0.1",
        "Hostname": "Host-1234",
        "Stats": [
            {
                "Start": 0,
                "Duration": 100000,
                "Stats": [
                    {
                        "Name": "http.request",
                        "Resource": "/user/profile",
                        "Service": "web-svc",
                        "Synthetics": False,
                        "Hits": ok_n + err_n,
                        "TopLevelHits": ok_n + err_n,
                        "Duration": total,
                        "Errors": err_n,
                        "OkSummary": DDSketchProto.to_proto(ok_dist).SerializeToString(),
                        "ErrorSummary": DDSketchProto.to_proto(err_dist).SerializeToString(),
                    },
                ],
            }
        ],
    }
    yield data


@pytest.fixture
def v06_reference_http_stats_payload_data(v06_reference_http_stats_payload_data_raw):
    yield msgpack.packb(v06_reference_http_stats_payload_data_raw)


@pytest.fixture
def do_reference_v06_http_stats(
    agent,
    v06_reference_http_stats_payload_headers,
    v06_reference_http_stats_payload_data,
):
    def fn(token: Optional[str] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.put(  # type: ignore
            "/v0.6/stats",
            params=params,
            headers=v06_reference_http_stats_payload_headers,
            data=v06_reference_http_stats_payload_data,
        )

    yield fn


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_data_raw():
    data = {
        "tracer_time": 1658439039,
        "runtime_id": "3cac6e9599564813977aace04bf37d57",
        "api_version": "v1",
        "seq_id": 1,
        "application": {
            "service_name": "my-svc",
            "service_version": "1.0.0",
            "env": "prod",
            "language_name": "python",
            "language_version": "3.9.10",
            "tracer_version": "1.3.0",
            "runtime_name": "CPython",
            "runtime_version": "3.9.10",
        },
        "host": {
            "os": "macOS-12.4",
            "hostname": "HELLO-COMPUTER",
            "os_version": "12.4",
            "kernel_name": "Darwin",
            "kernel_release": "21.5.0",
            "kernel_version": "Darwin Kernel Version 21.5.0: Tue Apr 26 21:08:22 PDT 2022; root:xnu-8020.121.3~4/RELEASE_X86_64",
            "container_id": "",
        },
        "payload": {
            "dependencies": [
                {"name": "pyparsing", "version": "3.0.9"},
                {"name": "pytest-mock", "version": "3.8.2"},
                {"name": "setuptools", "version": "62.6.0"},
                {"name": "sortedcontainers", "version": "2.4.0"},
                {"name": "attrs", "version": "21.4.0"},
                {"name": "wheel", "version": "0.37.1"},
                {"name": "protobuf", "version": "4.21.2"},
                {"name": "packaging", "version": "21.3"},
                {"name": "tomli", "version": "2.0.1"},
                {"name": "msgpack", "version": "1.0.4"},
                {"name": "bytecode", "version": "0.13.0"},
                {"name": "pip", "version": "22.1.2"},
                {"name": "py", "version": "1.11.0"},
                {"name": "ddsketch", "version": "2.0.3"},
                {"name": "coverage", "version": "6.4.2"},
                {"name": "pytest-cov", "version": "3.0.0"},
                {"name": "iniconfig", "version": "1.1.1"},
                {"name": "py-cpuinfo", "version": "8.0.0"},
                {"name": "toml", "version": "0.10.2"},
                {"name": "pluggy", "version": "1.0.0"},
                {"name": "mock", "version": "4.0.3"},
                {"name": "six", "version": "1.16.0"},
                {"name": "opentracing", "version": "2.4.0"},
                {"name": "pytest", "version": "6.2.5"},
                {"name": "ddtrace", "version": "1.3.0"},
                {"name": "tenacity", "version": "8.0.1"},
                {"name": "hypothesis", "version": "6.45.0"},
            ],
            "integrations": [],
            "configurations": [],
        },
        "request_type": "app-started",
    }
    yield data


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_data(
    v2_reference_http_apmtelemetry_payload_data_raw,
):
    yield json.dumps(v2_reference_http_apmtelemetry_payload_data_raw)


@pytest.fixture
def v07_reference_http_remoteconfig_payload_data_raw():
    data = {
        "client": {
            "id": "test-id",
            "products": ["ASM_DATA", "ASM", "ASM_DD"],
            "is_tracer": True,
            "client_tracer": {
                "runtime_id": "test_runtime_id",
                "language": "python",
                "tracer_version": "test-1.x",
                "service": "test_agent_service",
                "env": "staging",
                "app_version": "0.1",
                "tags": ["env:staging", "version:0.1", "tracer_version:test-1.x", "host_name:test_agent_service"],
            },
            "state": {
                "root_version": 1,
                "targets_version": 46924638,
                "config_states": [
                    {"id": "blocking", "version": 296, "product": "ASM", "apply_state": 2},
                    {"id": "custom_rules", "version": 24, "product": "ASM", "apply_state": 2},
                    {"id": "disabled_rules", "version": 1, "product": "ASM", "apply_state": 2},
                    {"id": "exclusion_filters", "version": 256, "product": "ASM", "apply_state": 2},
                    {"id": "blocked_ips", "version": 4732, "product": "ASM_DATA", "apply_state": 2},
                    {"id": "blocked_users", "version": 5, "product": "ASM_DATA", "apply_state": 2},
                    {"id": "16.recommended.json", "version": 1, "product": "ASM_DD", "apply_state": 2},
                ],
                "has_error": False,
                "backend_client_state": "backend_client_state_hash",
            },
            "capabilities": "Afw=",
        },
        "cached_target_files": [
            {
                "path": "datadog/2/ASM/blocking/config",
                "length": 21,
                "hashes": [
                    {"algorithm": "sha256", "hash": "d75be3984bbd00451d1d56eff464d63a05727f192f0fa367d1aac03869b5d74c"}
                ],
            },
            {
                "path": "datadog/2/ASM/custom_rules/config",
                "length": 425,
                "hashes": [
                    {"algorithm": "sha256", "hash": "5c49e86c48d88e3e836e2cb85469ebf05c02a5d1bc768639e4c45c75620deb23"}
                ],
            },
            {
                "path": "datadog/2/ASM/disabled_rules/config",
                "length": 57,
                "hashes": [
                    {"algorithm": "sha256", "hash": "ab06f434cf8b9bc50bf75c27abada0781dfb2590aea9c19963b86e87e236fe1b"}
                ],
            },
            {
                "path": "datadog/2/ASM/exclusion_filters/config",
                "length": 21,
                "hashes": [
                    {"algorithm": "sha256", "hash": "d75be3984bbd00451d1d56eff464d63a05727f192f0fa367d1aac03869b5d74c"}
                ],
            },
            {
                "path": "datadog/2/ASM_DATA/blocked_ips/config",
                "length": 1290,
                "hashes": [
                    {"algorithm": "sha256", "hash": "a5e3b56034a6ad61d9853cd1b80061d04068ada2d5e3b1b120a68c861006db27"}
                ],
            },
            {
                "path": "datadog/2/ASM_DATA/blocked_users/config",
                "length": 760,
                "hashes": [
                    {"algorithm": "sha256", "hash": "53e2dd0da7a7d6ca57f10c0e1d17fdfc7e44a987bb128265258b3bc7d0c48193"}
                ],
            },
            {
                "path": "employee/ASM_DD/16.recommended.json/config",
                "length": 203383,
                "hashes": [
                    {"algorithm": "sha256", "hash": "ad984f3d9020ee932783ee6448cc4efc8a58aa2c8d3f4fd8d2d730452a8dd4df"}
                ],
            },
        ],
    }
    yield data


@pytest.fixture
def v07_reference_http_remoteconfig_payload_data(
    v07_reference_http_remoteconfig_payload_data_raw,
):
    yield json.dumps(v07_reference_http_remoteconfig_payload_data_raw)


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_headers(  # type: ignore
    v2_reference_http_apmtelemetry_payload_data_raw,
) -> Generator[Dict[str, str], None, None]:
    headers = {
        "Content-type": "application/json",
        "DD-Telemetry-Request-Type": v2_reference_http_apmtelemetry_payload_data_raw["request_type"],
        "DD-Telemetry-API-Version": "v1",
    }
    yield headers


def v2_apmtelemetry(  # type: ignore
    agent,
    event: TelemetryEvent,
    token: Optional[str] = None,
):
    params = {"test_session_token": token} if token is not None else {}
    headers = {
        "Content-type": "application/json",
        "DD-Telemetry-Request-Type": event["request_type"],
        "DD-Telemetry-API-Version": "v1",
    }

    return agent.post(
        "/telemetry/proxy/api/v2/apmtelemetry",
        params=params,
        headers=headers,
        data=json.dumps(event),
    )


@pytest.fixture
def do_reference_v2_http_apmtelemetry(
    agent,
    v2_reference_http_apmtelemetry_payload_headers,
    v2_reference_http_apmtelemetry_payload_data,
):
    def fn(token: Optional[str] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.post(  # type: ignore
            "/telemetry/proxy/api/v2/apmtelemetry",
            params=params,
            headers=v2_reference_http_apmtelemetry_payload_headers,
            data=v2_reference_http_apmtelemetry_payload_data,
        )

    yield fn


@pytest.fixture
def available_port() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))  # Bind to a free port provided by the host.
    port = s.getsockname()[1]  # Get the port number assigned.
    s.close()  # Release the socket.
    return str(port)


@pytest.fixture
def testagent_port(available_port: str) -> str:
    return available_port


@pytest.fixture
def testagent_url(testagent_port: str) -> str:
    return "http://127.0.0.1:%s" % testagent_port


@pytest.fixture(scope="module")
def testagent_snapshot_ci_mode() -> bool:
    # Default all tests in this module to be run in CI mode
    # unless a special env var is passed to make generating
    # the snapshots easier.
    return os.getenv("GENERATE_SNAPSHOTS") != "1"


@pytest.fixture
def testagent_connection_type():
    return "http"


@pytest.fixture
def testagent_uds_socket_path(tmp_path):
    return tmp_path / "apm.socket"


@pytest.fixture
def test_agent_env(testagent_connection_type, testagent_uds_socket_path):
    env = os.environ.copy()
    if testagent_connection_type == "uds":
        env["DD_APM_RECEIVER_SOCKET"] = str(testagent_uds_socket_path)
    return env


@pytest.fixture
async def testagent(
    loop,
    testagent_port,
    testagent_snapshot_ci_mode,
    test_agent_env,
    testagent_connection_type,
    testagent_uds_socket_path,
):
    test_agent_env.update(
        {
            "PORT": testagent_port,
            "SNAPSHOT_CI": "1" if testagent_snapshot_ci_mode else "0",
            "SNAPSHOT_DIR": os.path.join(os.path.dirname(__file__), "integration_snapshots"),
        }
    )
    p = subprocess.Popen(["ddapm-test-agent"], env=test_agent_env)

    try:
        if testagent_connection_type == "uds":
            import time

            for i in range(100):
                if testagent_uds_socket_path.exists():
                    break
                time.sleep(0.05)
            else:
                raise AssertionError("Test agent did not create UDS socket in time")

            connector = aiohttp.UnixConnector(path=str(testagent_uds_socket_path))
            session = aiohttp.ClientSession(connector=connector)
            base_url = "http://localhost/"
        else:
            session = aiohttp.ClientSession()
            base_url = f"http://localhost:{testagent_port}"

        async with session:
            for _ in range(100):
                try:
                    r = await session.get(base_url)
                except (ClientConnectorError, ClientOSError):
                    pass
                else:
                    if r.status == 404:
                        break
                await asyncio.sleep(0.05)
            else:
                raise AssertionError("Test agent did not start in time")

            yield session
    finally:
        p.terminate()
        p.wait()


# OTLP Constants
PROTOBUF_HEADERS = {"Content-Type": "application/x-protobuf"}
JSON_HEADERS = {"Content-Type": "application/json"}


# OTLP Helper Functions
async def _get_http_status_from_metadata(call):
    """Extract HTTP status from GRPC trailing metadata."""
    trailing_metadata = await call.trailing_metadata()
    for key, value in trailing_metadata:
        if key == "http-status":
            return int(value)
    return None


# Common OTLP Test Data Fixtures
@pytest.fixture
def service_name():
    return "ddservice"


@pytest.fixture
def environment():
    return "ddenv"


@pytest.fixture
def version():
    return "ddv1"


@pytest.fixture
def host_name():
    return "ddhost"


# OTLP Infrastructure Fixtures
@pytest.fixture
def otlp_http_url(testagent_url):
    parsed_url = urlparse(testagent_url)
    return f"{parsed_url.scheme}://{parsed_url.hostname}:{DEFAULT_OTLP_HTTP_PORT}"


@pytest.fixture
def otlp_test_client(otlp_http_url):
    """OTLP client for retrieving stored logs, metrics and requests."""
    parsed_url = urlparse(otlp_http_url)
    client = TestOTLPClient(parsed_url.hostname, parsed_url.port, parsed_url.scheme)
    client.clear()
    client.wait_to_start()
    yield client
    client.clear()


@pytest.fixture(params=["logs", "metrics"])
async def otlp_grpc_client(request):
    """GRPC client that can connect to either logs or metrics service."""
    channel = grpc_aio.insecure_channel(f"127.0.0.1:{DEFAULT_OTLP_GRPC_PORT}")

    if request.param == "logs":
        stub = LogsServiceStub(channel)
    elif request.param == "metrics":
        stub = MetricsServiceStub(channel)
    else:
        raise ValueError(f"Unknown service type: {request.param}")

    yield stub

    await channel.close()


@pytest.fixture
async def otlp_logs_grpc_client():
    """GRPC client specifically for logs service."""
    channel = grpc_aio.insecure_channel(f"127.0.0.1:{DEFAULT_OTLP_GRPC_PORT}")
    stub = LogsServiceStub(channel)

    yield stub

    await channel.close()


@pytest.fixture
async def otlp_metrics_grpc_client():
    """GRPC client specifically for metrics service."""
    channel = grpc_aio.insecure_channel(f"127.0.0.1:{DEFAULT_OTLP_GRPC_PORT}")
    stub = MetricsServiceStub(channel)

    yield stub

    await channel.close()


@pytest.fixture
async def grpc_client_with_failure_type(agent_app, available_port, aiohttp_server, request):
    """GRPC server with configurable HTTP backend failure scenarios for both logs and metrics."""
    grpc_port = int(available_port)

    # Require explicit tuple of (failure_type, service_type)
    param = getattr(request, "param")
    if not isinstance(param, tuple) or len(param) != 2:
        raise ValueError("grpc_client_with_failure_type requires a tuple of (failure_type, service_type)")

    failure_type, service_type = param
    if service_type not in ["logs", "metrics"]:
        raise ValueError(f"service_type must be 'logs' or 'metrics', got: {service_type}")

    endpoint = LOGS_ENDPOINT if service_type == "logs" else METRICS_ENDPOINT

    http_handlers = {
        "http_400": lambda _: web.HTTPBadRequest(text="invalid"),
        "http_500": lambda _: web.HTTPInternalServerError(text="boom"),
    }

    if failure_type == "connection_failure":
        http_port = 99999  # Non-existent port
    elif failure_type in http_handlers:
        app = web.Application()
        app.router.add_post(endpoint, http_handlers[failure_type])
        http_server = await aiohttp_server(app)
        http_port = http_server.port
    else:
        raise ValueError(f"Unknown failure_type: {failure_type}")

    grpc_server = await make_otlp_grpc_server_async(agent_app.app["agent"], http_port=http_port, grpc_port=grpc_port)
    channel = grpc_aio.insecure_channel(f"localhost:{grpc_port}")

    if service_type == "logs":
        stub = LogsServiceStub(channel)
    elif service_type == "metrics":
        stub = MetricsServiceStub(channel)
    else:
        raise ValueError(f"Unknown service_type: {service_type}")

    yield stub

    await channel.close()
    await grpc_server.stop(grace=0)



"""OTLP Metrics handling for the test agent."""

import json
import logging
from typing import Any
from typing import Dict

from aiohttp import ClientSession
from google.protobuf.json_format import MessageToDict
from grpc import aio as grpc_aio
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import ExportMetricsServiceRequest
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import ExportMetricsServiceResponse
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2_grpc import MetricsServiceServicer


METRICS_ENDPOINT = "/v1/metrics"


log = logging.getLogger(__name__)


def decode_metrics_request(request_body: bytes, content_type: str) -> Dict[str, Any]:
    if content_type == "application/json":
        parsed_json = json.loads(request_body)
        if not isinstance(parsed_json, dict):
            raise Exception("JSON payload must be an object")
        return parsed_json
    elif content_type == "application/x-protobuf":
        export_request = ExportMetricsServiceRequest()
        export_request.ParseFromString(request_body)
        return protobuf_to_dict(export_request)
    else:
        raise ValueError(f"Content-Type must be application/x-protobuf or application/json, got {content_type}")


def protobuf_to_dict(pb_obj: Any) -> Dict[str, Any]:
    result: Dict[str, Any] = MessageToDict(pb_obj, preserving_proto_field_name=True)
    return result


class OTLPMetricsGRPCServicer(MetricsServiceServicer):

    def __init__(self, http_port: int):
        self.http_url = f"http://127.0.0.1:{http_port}"

    def _count_data_points(self, request: ExportMetricsServiceRequest) -> int:
        return len(
            [
                dp
                for rm in request.resource_metrics
                for sm in rm.scope_metrics
                for m in sm.metrics
                for dp in (
                    m.gauge.data_points
                    if m.HasField("gauge")
                    else (
                        m.sum.data_points
                        if m.HasField("sum")
                        else (
                            m.histogram.data_points
                            if m.HasField("histogram")
                            else (
                                m.exponential_histogram.data_points
                                if m.HasField("exponential_histogram")
                                else m.summary.data_points if m.HasField("summary") else []
                            )
                        )
                    )
                )
            ]
        )

    async def Export(
        self, request: ExportMetricsServiceRequest, context: grpc_aio.ServicerContext
    ) -> ExportMetricsServiceResponse:
        try:
            protobuf_data = request.SerializeToString()
            headers = {"Content-Type": "application/x-protobuf"}
            metadata = dict(context.invocation_metadata())
            if "session-token" in metadata:
                headers["Session-Token"] = metadata["session-token"]
            async with ClientSession(self.http_url) as session:
                async with session.post(METRICS_ENDPOINT, headers=headers, data=protobuf_data) as resp:
                    context.set_trailing_metadata([("http-status", str(resp.status))])
                    response = ExportMetricsServiceResponse()
                    if resp.status >= 400:
                        response.partial_success.rejected_data_points = self._count_data_points(request)
                        response.partial_success.error_message = f"HTTP {resp.status}: {await resp.text()}"
                    return response
        except Exception as e:
            context.set_trailing_metadata([("http-status", "500"), ("error", str(e))])
            response = ExportMetricsServiceResponse()
            response.partial_success.rejected_data_points = self._count_data_points(request)
            response.partial_success.error_message = f"Forward failed: {str(e)}"
            return response



import base64
import json

from google.protobuf.json_format import MessageToDict
from opentelemetry.proto.collector.logs.v1.logs_service_pb2 import ExportLogsServiceRequest
from opentelemetry.proto.collector.logs.v1.logs_service_pb2 import ExportLogsServiceResponse
from opentelemetry.proto.common.v1.common_pb2 import AnyValue
from opentelemetry.proto.common.v1.common_pb2 import KeyValue
from opentelemetry.proto.logs.v1.logs_pb2 import LogRecord
from opentelemetry.proto.logs.v1.logs_pb2 import ResourceLogs
from opentelemetry.proto.logs.v1.logs_pb2 import ScopeLogs
from opentelemetry.proto.resource.v1.resource_pb2 import Resource
import pytest

from ddapm_test_agent.logs import LOGS_ENDPOINT
from tests.conftest import JSON_HEADERS
from tests.conftest import PROTOBUF_HEADERS
from tests.conftest import _get_http_status_from_metadata


def _find_service_name_in_resource(resource_logs, expected_service_name):
    if not resource_logs or not resource_logs[0].get("resource"):
        return False

    resource = resource_logs[0]["resource"]
    for attr in resource.get("attributes", []):
        if attr.get("key") == "service.name" and attr.get("value", {}).get("string_value") == expected_service_name:
            return True
    return False


@pytest.fixture
def log_message():
    return "test_otel_logs_exporter_auto_configured_http"


@pytest.fixture
def trace_id():
    return b""


@pytest.fixture
def span_id():
    return b""


@pytest.fixture
def otlp_logs_protobuf(service_name, environment, version, host_name, log_message, trace_id, span_id):
    resource = Resource()
    resource.attributes.extend(
        [
            KeyValue(key="service.name", value=AnyValue(string_value=service_name)),
            KeyValue(key="deployment.environment.name", value=AnyValue(string_value=environment)),
            KeyValue(key="service.version", value=AnyValue(string_value=version)),
            KeyValue(key="host.name", value=AnyValue(string_value=host_name)),
        ]
    )

    log_record = LogRecord()
    log_record.body.string_value = log_message
    log_record.trace_id = trace_id
    log_record.span_id = span_id

    scope_logs = ScopeLogs()
    scope_logs.log_records.append(log_record)

    resource_logs = ResourceLogs()
    resource_logs.resource.CopyFrom(resource)
    resource_logs.scope_logs.append(scope_logs)

    export_request = ExportLogsServiceRequest()
    export_request.resource_logs.append(resource_logs)

    return export_request


@pytest.fixture
def otlp_logs_string(otlp_logs_protobuf):
    return otlp_logs_protobuf.SerializeToString()


@pytest.fixture
def otlp_logs_json(otlp_logs_protobuf):
    return json.dumps(MessageToDict(otlp_logs_protobuf, preserving_proto_field_name=True))


async def test_logs_endpoint_basic_http(testagent, otlp_http_url, otlp_logs_string, loop):
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200
    assert resp.content_type == "application/x-protobuf"
    body = await resp.read()
    assert body == ExportLogsServiceResponse().SerializeToString()


@pytest.mark.parametrize(
    "service_name,environment,version,host_name,log_message,trace_id,span_id",
    [
        (
            "web-service",
            "production",
            "1.0.0",
            "web-01",
            "User login successful",
            b"\xdd\xda\xa1\xa9\xa5y\xc1\xbd",
            b"h\x9b[\xe5\x00\x00\x00\x00\x80\x1b\\]\x1c#\xe6\xfb",
        ),
        (
            "api-service",
            "staging",
            "2.1.3",
            "api-02",
            "Database connection established",
            b"h\xb8\xcf\xb9\xa5;\x0f\x87",
            b"h\x9b[\xe5\x00\x00\x00\x00\x83\x0f\xb8M\xc8\xac\xd6Z",
        ),
        (
            "payment-service",
            "development",
            "0.9.5",
            "payment-03",
            "Payment processed successfully",
            b"\xee\x1fl\xf8d\x9a\xc2+",
            b"h\x9b[\xe5\x00\x00\x00\x00v{\xd8\xfd5+K(",
        ),
        ("auth-service", "test", "3.2.1", "auth-04", "Token validation failed", b"", b""),  # Empty trace/span IDs
        (
            "notification-service",
            "production",
            "1.5.7",
            "notify-05",
            "Email sent to user@example.com",
            b"\xb0\x10)\xbaI\xbc\x9d\xf1",
            b"h\x9b[\xe5\x00\x00\x00\x00k_$\x95`\n\xbf\xc5",
        ),
    ],
)
async def test_session_logs_endpoint_http(
    testagent,
    otlp_http_url,
    otlp_logs_string,
    service_name,
    environment,
    version,
    host_name,
    log_message,
    span_id,
    trace_id,
    loop,
):
    """Session endpoint returns logs with all attributes preserved."""
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200

    resp = await testagent.get(f"{otlp_http_url}/test/session/logs")
    assert resp.status == 200
    logs = await resp.json()
    assert len(logs) == 1
    assert "resource_logs" in logs[0]

    resource_logs = logs[0]["resource_logs"]
    assert len(resource_logs) == 1

    assert _find_service_name_in_resource(
        resource_logs, service_name
    ), f"service.name should be set to '{service_name}' in resource attributes"

    assert len(resource_logs) == 1
    resource = resource_logs[0].get("resource", {})
    assert resource.get("attributes") == [
        {"key": "service.name", "value": {"string_value": service_name}},
        {"key": "deployment.environment.name", "value": {"string_value": environment}},
        {"key": "service.version", "value": {"string_value": version}},
        {"key": "host.name", "value": {"string_value": host_name}},
    ]
    scope_logs = resource_logs[0].get("scope_logs", [])
    assert len(scope_logs) == 1
    log_records = scope_logs[0]["log_records"]
    assert len(log_records) == 1
    assert log_records[0]["body"].get("string_value") == log_message
    # trace_id and span_id are stored as base64 encoded strings in JSON
    expected_trace_id = base64.b64encode(trace_id).decode("ascii") if trace_id else None
    expected_span_id = base64.b64encode(span_id).decode("ascii") if span_id else None
    assert log_records[0].get("trace_id") == expected_trace_id
    assert log_records[0].get("span_id") == expected_span_id


async def test_otlp_client_logs(testagent, otlp_test_client, otlp_http_url, otlp_logs_string, loop):
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200

    otlp_test_client.wait_for_num_logs(1)

    resp = otlp_test_client.requests()
    assert len(resp) == 1
    assert resp[0]["method"] == "POST"
    assert resp[0]["url"] == f"{otlp_http_url}{LOGS_ENDPOINT}"
    assert resp[0]["headers"]["Content-Type"] == PROTOBUF_HEADERS["Content-Type"]
    decoded_body = base64.b64decode(resp[0]["body"])
    assert (
        decoded_body == otlp_logs_string
    ), f"body: {resp[0]['body']} decoded: {decoded_body}, otlp_logs_string: {otlp_logs_string}"

    logs = otlp_test_client.logs()
    assert len(logs) == 1
    assert "resource_logs" in logs[0]

    otlp_test_client.clear()
    logs = otlp_test_client.logs()
    assert len(logs) == 0


async def test_logs_endpoint_integration_http(
    testagent, otlp_http_url, otlp_logs_string, service_name, log_message, loop
):
    resp = await testagent.get(f"{otlp_http_url}/test/session/clear")
    assert resp.status == 200

    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200

    resp = await testagent.get(f"{otlp_http_url}/test/session/logs")
    assert resp.status == 200
    captured_logs_list = await resp.json()

    assert len(captured_logs_list) > 0, "Expected at least one resource log"

    captured_logs = captured_logs_list[0]
    resource_logs = captured_logs["resource_logs"]
    assert len(resource_logs) > 0

    assert _find_service_name_in_resource(
        resource_logs, service_name
    ), f"service.name should be set to '{service_name}' in resource attributes"

    scope_logs = resource_logs[0].get("scope_logs", [])
    assert len(scope_logs) == 1
    log_records = scope_logs[0]["log_records"]
    assert len(log_records) == 1
    assert log_records[0]["body"].get("string_value") == log_message


async def test_multiple_logs_sessions_http(testagent, otlp_http_url, otlp_logs_string, loop):
    """Logs are isolated between sessions."""
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200

    resp = await testagent.get(f"{otlp_http_url}/test/session/start")
    assert resp.status == 200

    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=PROTOBUF_HEADERS, data=otlp_logs_string)
    assert resp.status == 200

    resp = await testagent.get(f"{otlp_http_url}/test/session/logs")
    assert resp.status == 200
    logs = await resp.json()
    assert len(logs) == 1  # Only the log from the current session


async def test_logs_endpoint_json_http(testagent, otlp_http_url, otlp_logs_json, service_name, loop):
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=JSON_HEADERS, data=otlp_logs_json)
    assert resp.status == 200

    resp = await testagent.get(f"{otlp_http_url}/test/session/logs")
    assert resp.status == 200
    logs = await resp.json()
    assert len(logs) == 1

    resource_logs = logs[0]["resource_logs"]
    assert _find_service_name_in_resource(
        resource_logs, service_name
    ), f"service.name should be set to '{service_name}' in resource attributes"


async def test_logs_endpoint_invalid_content_type(testagent, otlp_http_url, otlp_logs_string, loop):
    resp = await testagent.post(
        f"{otlp_http_url}{LOGS_ENDPOINT}", headers={"Content-Type": "application/xml"}, data=otlp_logs_string
    )
    assert resp.status == 400

    resp = await testagent.post(
        f"{otlp_http_url}{LOGS_ENDPOINT}", headers={"Content-Type": "text/plain"}, data=b"some plain text"
    )
    assert resp.status == 400

    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", data=otlp_logs_string)
    assert resp.status == 400


async def test_logs_endpoint_invalid_json(testagent, otlp_http_url, loop):
    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=JSON_HEADERS, data=b'{"invalid": json}')
    assert resp.status == 400

    resp = await testagent.post(
        f"{otlp_http_url}{LOGS_ENDPOINT}", headers=JSON_HEADERS, data=b'["not", "an", "object"]'
    )
    assert resp.status == 400

    resp = await testagent.post(f"{otlp_http_url}{LOGS_ENDPOINT}", headers=JSON_HEADERS, data=b'"just a string"')
    assert resp.status == 400


async def test_logs_endpoint_basic_grpc(testagent, otlp_logs_grpc_client, otlp_logs_protobuf, loop):
    call = otlp_logs_grpc_client.Export(otlp_logs_protobuf)
    response = await call

    assert response is not None

    http_status = await _get_http_status_from_metadata(call)
    assert http_status == 200, f"Expected HTTP 200, got {http_status}"

    # For successful requests, partial_success should be empty
    assert response.partial_success.rejected_log_records == 0
    assert response.partial_success.error_message == ""


async def test_session_logs_endpoint_grpc_forwarding(
    testagent, otlp_logs_grpc_client, otlp_test_client, otlp_logs_protobuf, service_name, log_message, loop
):
    call = otlp_logs_grpc_client.Export(otlp_logs_protobuf)
    response = await call
    assert response is not None

    http_status = await _get_http_status_from_metadata(call)
    assert http_status == 200, f"Expected HTTP 200, got {http_status}"

    logs = otlp_test_client.logs()

    resource_logs = logs[0]["resource_logs"]
    assert _find_service_name_in_resource(
        resource_logs, service_name
    ), f"service.name should be set to '{service_name}' in resource attributes"

    scope_logs = resource_logs[0].get("scope_logs", [])
    assert len(scope_logs) == 1
    log_records = scope_logs[0]["log_records"]
    assert len(log_records) == 1
    assert log_records[0]["body"].get("string_value") == log_message


@pytest.mark.parametrize("grpc_client_with_failure_type", [("http_400", "logs")], indirect=True)
async def test_grpc_maps_http_400_to_metadata(grpc_client_with_failure_type):
    """GRPC forwarding preserves HTTP 400 status in metadata and partial_success."""
    call = grpc_client_with_failure_type.Export(ExportLogsServiceRequest())
    response = await call
    assert response is not None

    http_status = await _get_http_status_from_metadata(call)
    assert http_status == 400, f"Expected HTTP 400, got {http_status}"

    assert response.partial_success.rejected_log_records == 0  # Empty request
    assert "HTTP 400" in response.partial_success.error_message


@pytest.mark.parametrize("grpc_client_with_failure_type", [("http_500", "logs")], indirect=True)
async def test_grpc_maps_http_500_to_metadata(grpc_client_with_failure_type):
    """GRPC forwarding preserves HTTP 500 status in metadata and partial_success."""
    call = grpc_client_with_failure_type.Export(ExportLogsServiceRequest())
    response = await call
    assert response is not None

    http_status = await _get_http_status_from_metadata(call)
    assert http_status == 500, f"Expected HTTP 500, got {http_status}"

    assert response.partial_success.rejected_log_records == 0  # Empty request
    assert "HTTP 500" in response.partial_success.error_message


@pytest.mark.parametrize("grpc_client_with_failure_type", [("connection_failure", "logs")], indirect=True)
async def test_grpc_server_resilience_after_failure(grpc_client_with_failure_type, otlp_logs_protobuf):
    """GRPC server remains operational after processing failed requests."""
    call1 = grpc_client_with_failure_type.Export(otlp_logs_protobuf)
    response1 = await call1
    assert response1 is not None

    assert response1.partial_success.rejected_log_records == 1  # 1 ResourceLogs object
    assert "Forward failed" in response1.partial_success.error_message

    http_status = await _get_http_status_from_metadata(call1)
    assert http_status == 500  # Connection failure mapped to 500

    call2 = grpc_client_with_failure_type.Export(otlp_logs_protobuf)
    response2 = await call2
    assert response2 is not None

    assert response2.partial_success.rejected_log_records == 1  # 1 ResourceLogs object
    assert "Forward failed" in response2.partial_success.error_message

    call3 = grpc_client_with_failure_type.Export(ExportLogsServiceRequest())
    response3 = await call3
    assert response3 is not None
